/***********************************************************************
 * Bdbj 1.2 (by Jon Cook May 2002)
 *
 * Watches.java
 * 
 * Maintain a visible watch table, send watch queries to the client,
 * and update values shown on receiving reply. 
 ***********************************************************************/

package jjc.bdbj.interfaces.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;

public class Watches implements ActionListener {
    private WatchTable watchTable;
    private JScrollPane watchTableScrollPane = null;
    private MessageReceiver messageReceiver = null;

    public Watches( ) {
        watchTable = new WatchTable( );
        watchTable.setFont( new Font( "Monospaced", Font.PLAIN, 14 ) );
        watchTableScrollPane = new JScrollPane( watchTable );
    }   
  
    public void setMessageReceiver( MessageReceiver messageReceiver ) {
	watchTable.setMessageReceiver( messageReceiver );
    }

    public JScrollPane getWatchTableScrollPane( ) {
        return watchTableScrollPane;
    }

    public void reportWatch( String watchReport ) {
        // watchReport is of the form <identifier> == <value>
        int separatorLocation = watchReport.indexOf( " == " );
        String identifier = watchReport.substring( 0, separatorLocation ).trim( );
        String value = watchReport.substring( separatorLocation + 4 );
        watchTable.reportWatch( identifier, value );
    }

    public void triggerFullWatchUpdate( ) {
        watchTable.triggerFullWatchUpdate( );
    }

    private void insertWatch( ) {
	watchTable.insertRow( );
    }

    private void deleteWatch( ) {
        watchTable.removeRow( );
    }

    private void deleteAllWatches( ) {
        watchTable.removeAllRows( );
    }

    public MenuData getMenu( ) {
        boolean[] bothModes = { true, true, true };
        MenuItemData[] watchMenuData =
            { new MenuItemData( "Delete Selected Watches", "Delete Selected Watches", this, "alt 9", bothModes ),
	      new MenuItemData( "Insert Watch Field", "Insert Watch Field", this, "alt shift 9", bothModes ),
              new MenuItemData( "Delete All Watches", "Delete All Watches", this, "alt 0", bothModes ) }; 
        return new MenuData( "Watches", watchMenuData );
    }

    public void actionPerformed( ActionEvent event ) {
        String label = event.getActionCommand( );
        if( label.equals( "Delete Selected Watches" ) ) 
            deleteWatch( );
	else if( label.equals( "Insert Watch Field" ) )
	    insertWatch( );
        else if( label.equals( "Delete All Watches" ) )
            deleteAllWatches( );
    }
}

class WatchTable extends JTable implements TableModelListener {
    private DefaultTableModel tableModel = null;
    private TableColumnModel columnModel = null;
    private TableColumn varColumn = null;
    private TableColumn valColumn = null;
    private JTextField textField = null;    
    private JTextField viewField = null;

    private MessageReceiver messageReceiver = null;
    public void setMessageReceiver( MessageReceiver messageReceiver ) {
	this.messageReceiver = messageReceiver;
    }

    Object[] titles = { "Variable", "Value" };

    public WatchTable( ) {
        super( );
        setModel( tableModel = new WatchTableModel( titles, 2 ) );
	setSelectionMode( ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );
        setColumnModel( columnModel = new DefaultTableColumnModel( ) );

        columnModel.addColumn( varColumn = new TableColumn( 0 ) );
        varColumn.setHeaderValue( "Variable" );
        textField = new JTextField( );
        textField.setFont(new Font( "Monospaced", Font.PLAIN, 14 ) );
        varColumn.setCellEditor( new DefaultCellEditor( textField ) );

        columnModel.addColumn( valColumn = new TableColumn( 1 ) );
        valColumn.setHeaderValue( "Value" );
	valColumn.setCellRenderer( new RedRenderer( ) );
        viewField = new JTextField( );
        viewField.setEditable( false );
        viewField.setFont(new Font( "Monospaced", Font.PLAIN, 14 ) );
        valColumn.setCellEditor( new DefaultCellEditor( viewField ) );

        tableModel.addTableModelListener( this );
    }

    public void reportWatch( String identifier, String value ) {
        int rows = tableModel.getRowCount( );
        for( int i = 0; i < rows; i++ ) 
	    if( tableModel.getValueAt( i, 0 ) != null )
		if( tableModel.getValueAt( i, 0 ).toString( ).trim( ).equals( identifier ) )
		    tableModel.setValueAt( value, i, 1 );
	repaint( );
    }

    public synchronized void triggerFullWatchUpdate( ) {
        int rows = tableModel.getRowCount( );
        for( int i = 0; i < rows; i++ )
            getWatch( tableModel.getValueAt( i, 0 ) );
    }

    public void insertRow( ) {
	editingStopped( new ChangeEvent( this ) );
	int selectedRow = getSelectedRow( );
	if( selectedRow != -1 )
	    tableModel.insertRow( selectedRow, nullRow );
	revalidate( );
	repaint( );
    }

    public void removeRow( ) { // now misnomer.
	editingStopped( new ChangeEvent( this ) );
	int[] selectedRows = getSelectedRows( );
	for( int i = selectedRows.length - 1; i >= 0; i-- ) {
	    if( selectedRows[ i ] < getRowCount( ) - 2 )
		tableModel.removeRow( selectedRows[ i ] );
	    else {
		tableModel.setValueAt( null, selectedRows[ i ], 0 );
		tableModel.setValueAt( null, selectedRows[ i ], 1 );
	    }
	}
	tableChanged( new TableModelEvent( tableModel ) );
	if( selectedRows.length > 1 )
	    clearSelection( );
	revalidate( );
	repaint( );
    }

    public void removeAllRows( ) {
	editingStopped( new ChangeEvent( this ) );
        int rows = tableModel.getRowCount( );
        for( int i = 0; i < rows; i++ )
            tableModel.removeRow( 0 );
        tableModel.addRow( nullRow );
        tableModel.addRow( nullRow );
        tableChanged( new TableModelEvent( tableModel ) );
	clearSelection( );
	revalidate( );
        repaint( );
    }

    private Object[] nullRow = { null, null };

    public void getWatch( Object identifier ) {
        if( identifier == null )
            return;
        if( identifier.toString( ).equals( "" ) ) {
	    reportWatch( "", "" );
            return;
	}

	/* the equals sign is used internally, and unless we take
	 * special precautions the user would be able to cause bizarre
	 * values to be displayed if a watch request were sent
	 * containing an equals sign. */
	if( identifier.toString( ).indexOf( "==" ) != -1 )
	    reportWatch( identifier.toString( ), "<please do not use the double equals sign>" );
	else
	    messageReceiver.deliverMessage( "watch " + identifier.toString( ).trim( ) );
	
	/*
	  if( debug.menuBar.getCurrentMode( ).equals( "Debug" ) )
	  debug.send( "watch " + identifier.toString( ) );
	  else
	  reportWatch( identifier.toString( ), "<not debugging>" );
	*/
    }

    public void tableChanged( TableModelEvent event ) {
        if( tableModel == null )
            return;
        int rows = tableModel.getRowCount( );
        if( event.getType( ) == TableModelEvent.UPDATE &&
            event.getColumn( ) == 0 ) {

            // update the watch value for the changed row
            int changedRow = event.getFirstRow( );
            if( changedRow >= 0 ) {
		getWatch( getValueAt( changedRow, 0 ) );
                repaint( );
            }

            // if there is any text in either of the last two rows then add a row
            if( tableModel.getValueAt( rows - 2, 0 ) != null && 
                !tableModel.getValueAt( rows - 2, 0 ).toString( ).equals( "" ) ||
                tableModel.getValueAt( rows - 1, 0 ) != null &&
                !tableModel.getValueAt( rows - 1 , 0 ).toString( ).equals( "" ) ) {
                tableModel.addRow( nullRow );
                repaint( );
            }
        }
    }
}

class WatchTableModel extends DefaultTableModel implements TableModel {

    public WatchTableModel( Object[] titles, int rows ) {
        super( titles, rows );
    }

    public String getColumnName( int nCol ) {
        return nCol==0 ? "Variable" : "Value";
    }

    public synchronized Object getValueAt( int row, int column ) {
        return super.getValueAt( row, column );
    }

    public synchronized void setValueAt( Object value, int row, int column ) {
        super.setValueAt( value, row, column );
    }
}

class RedRenderer extends DefaultTableCellRenderer {
    public void setValue( Object value ) {
	if( value == null ) {
	    setText( "" );
	    return;
	}
	if( value.toString( ).length( ) >= 1 && value.toString( ).charAt( 0 ) == '<' )
	    setForeground( Color.red );
	else
	    setForeground( Color.black );
	setText( value.toString( ) );
    }
}
